function f = image_enhancement
f.normalization = @normalization;
f.fftenhance = @fftenhance;
f.adaptiveThres = @adaptiveThres;
f.direction = @direction;
f.drawROI = @drawROI;
f.thinning = @thinning;




function norm_image = normalization(image) 
norm_image = histeq(uint8(image));
imagesc(norm_image);

function [final]=fftenhance(image,f)
I = 255-double(image);
[w,h] = size(I);
w1=floor(w/32)*32;
h1=floor(h/32)*32;
inner = zeros(w1,h1);
for i=1:32:w1
   for j=1:32:h1
      a=i+31;
      b=j+31;
      F=fft2( I(i:a,j:b) );
      factor=abs(F).^f;
      block = abs(ifft2(F.*factor));
      larv=max(block(:));
      if larv==0
         larv=1;
      end;
      block= block./larv;
      inner(i:a,j:b) = block;
   end;
end;
final=inner*255;
final=histeq(uint8(final));

function [o] = adaptiveThres(a,W,noShow);
[w,h] = size(a);
o = zeros(w,h);
%seperate it to W block
%step to w with step length W
for i=1:W:w
    for j=1:W:h
        mean_thres = 0;
        if i+W-1 <= w & j+W-1 <= h
            mean_thres = mean2(a(i:i+W-1,j:j+W-1));
            %threshold value is choosed
            mean_thres = 0.8*mean_thres;
            %before binarization
            %ridges are black, small intensity value -> 1 (white ridge)
            %the background and valleys are white, large intensity value -> 0(black)
            o(i:i+W-1,j:j+W-1) = a(i:i+W-1,j:j+W-1) < mean_thres;
        end;
    end;
end;
if nargin == 2
imagesc(o);
colormap(gray);
end;


function [p,z] = direction(image,blocksize,noShow)
%	DIRECTION cacluates the local flow orientation in each local window
%  with size (blocksize x blocksize)
%  direction(grayScaleFingerprintImage,blocksize,graphicalShowDisableFlag)
%  return p ROI bound
%  return z ROI area
[w,h] = size(image);
direct = zeros(w,h);
gradient_times_value = zeros(w,h);
gradient_sq_minus_value = zeros(w,h);
gradient_for_bg_under = zeros(w,h);
W = blocksize;
theta = 0;
sum_value = 1;
bg_certainty = 0;
blockIndex = zeros(ceil(w/W),ceil(h/W));
times_value = 0;
minus_value = 0;
center = [];
%Note that the image coordinate system is
%x axis towards bottom and y axis towards right
filter_gradient = fspecial('sobel');
%to get x gradient
I_horizontal = filter2(filter_gradient,image);
%to get y gradient
filter_gradient = transpose(filter_gradient);
I_vertical = filter2(filter_gradient,image);
gradient_times_value=I_horizontal.*I_vertical;
gradient_sq_minus_value=(I_vertical-I_horizontal).*(I_vertical+I_horizontal);
gradient_for_bg_under = (I_horizontal.*I_horizontal) + (I_vertical.*I_vertical);
for i=1:W:w
    for j=1:W:h
      if j+W-1 < h & i+W-1 < w
          times_value = sum(sum(2*gradient_times_value(i:i+W-1, j:j+W-1)));
          minus_value = sum(sum(gradient_sq_minus_value(i:i+W-1, j:j+W-1)));
          sum_value = sum(sum(gradient_for_bg_under(i:i+W-1, j:j+W-1)));
          bg_certainty = 0;
          theta = 0;
          if sum_value ~= 0 & times_value ~=0
              bg_certainty = sqrt((times_value*times_value + minus_value*minus_value)/(W*W*sum_value));
              if bg_certainty > 0.05 
                  blockIndex(ceil(i/W),ceil(j/W)) = 1;
                  tan_value = atan2(2*times_value,minus_value);
                  theta = (tan_value)/2 ;
                  theta = theta+pi/2;
                  center = [center;[round(i + (W-1)/2),round(j + (W-1)/2),theta]];
              end;
          end;
      end;
      times_value = 0;
      minus_value = 0;
      sum_value = 0;
    end;
end;
if nargin == 2
	imagesc(direct);

	hold on;
	[u,v] = pol2cart(center(:,3),8);
   quiver(center(:,2),center(:,1),u,v,0,'r');
   hold off;
end;
x = bwlabel(blockIndex,4);
y = bwmorph(x,'close');
z = bwmorph(y,'open');
p = bwperim(z);

function [roiImg,roiBound,roiArea] = drawROI(in,inBound,inArea,noShow)
%  drawROI(grayLevelFingerprintImage,ROIboundMap,ROIareaMap,flagToDisableGUI)
%  construct a ROI rectangle for the input fingerprint image and return the 
%  segmented fingerprint
%  With the assumption that only one ROI region for each fingerprint image
[iw,ih]=size(in);
tmplate = zeros(iw,ih);
[w,h] = size(inArea);
tmp=zeros(iw,ih);
left = 1;
right = h;
upper = 1;
bottom = w;
le2ri = sum(inBound);
roiColumn = find(le2ri>0);
left = min(roiColumn);
right = max(roiColumn);
tr_bound = inBound';
up2dw=sum(tr_bound);
roiRow = find(up2dw>0);
upper = min(roiRow);
bottom = max(roiRow);
%cut out the ROI region image
%show background,bound,innerArea with different gray intensity:0,100,200
for i = upper:1:bottom
   for j = left:1:right
      if inBound(i,j) == 1
         tmplate(16*i-15:16*i,16*j-15:16*j) = 200;
         tmp(16*i-15:16*i,16*j-15:16*j) = 1;
      elseif inArea(i,j) == 1 & inBound(i,j) ~=1
      %if inArea(i,j) == 1 & inBound(i,j) ~=1
         tmplate(16*i-15:16*i,16*j-15:16*j) = 100;
         tmp(16*i-15:16*i,16*j-15:16*j) = 1;
         
      end;
   end;
end;
in=in.*tmp;
roiImg = in(16*upper-15:16*bottom,16*left-15:16*right);
roiBound = inBound(upper:bottom,left:right);
roiArea = inArea(upper:bottom,left:right);
roiArea = im2double(roiArea) - im2double(roiBound);
if nargin == 3
	colormap(gray);
   imagesc(tmplate);
end;

function image=thinning(im)
image = im2double(bwmorph(im,'thin',Inf));